# coding: utf-8

from __future__ import print_function
from __future__ import absolute_import

_package_data = dict(
    full_package_name="pon",
    version_info=(0, 1, 1),
    author="Anthon van der Neut",
    author_email="a.van.der.neut@ruamel.eu",
    description="Python Object Notation",
    keywords="pon object-notation json yaml xml",
    entry_points=None,    # delete or `my_cmd = pon:main` to get executable
    license="MIT license",
    # status=u"α",
    # data_files="",
    universal=1,
    install_requires=dict(
        any = [],
        py27=["ruamel.ordereddict"],
        py26=["ruamel.ordereddict"],
    ),
)   # NOQA


# < from ruamel.util.new import _convert_version
def _convert_version(tup):
    """Create a PEP 386 pseudo-format conformant string from tuple tup."""
    ret_val = str(tup[0])  # first is always digit
    next_sep = "."  # separator for next extension, can be "" or "."
    for x in tup[1:]:
        if isinstance(x, int):
            ret_val += next_sep + str(x)
            next_sep = '.'
            continue
        first_letter = x[0].lower()
        next_sep = ''
        if first_letter in 'abcr':
            ret_val += 'rc' if first_letter == 'r' else first_letter
        elif first_letter in 'pd':
            ret_val += '.post' if first_letter == 'p' else '.dev'
    return ret_val
# <

version_info = _package_data['version_info']
__version__ = _convert_version(version_info)

del _convert_version

import string                                            # NOQA
try:
    from ruamel.ordereddict import ordereddict           # NOQA
except ImportError:
    from collections import OrderedDict as ordereddict   # NOQA
import tokenize                                          # NOQA

import sys                                               # NOQA
import platform                                          # NOQA
import datetime                                          # NOQA
from textwrap import dedent                              # NOQA
from _ast import *                                       # NOQA

if sys.version_info < (3, ):
    string_type = basestring
else:
    string_type = str

if sys.version_info < (3, 4):
    class Bytes():
        pass

    class NameConstant:
        pass

if sys.version_info < (2, 7) or platform.python_implementation() == 'Jython':
    class Set():
        pass


def loads(node_or_string, dict_typ=dict, return_ast=False, file_name=None):
    """
    Safely evaluate an expression node or a string containing a Python
    expression.  The string or node provided may only consist of the following
    Python literal structures: strings, bytes, numbers, tuples, lists, dicts,
    sets, booleans, and None.
    """
    if sys.version_info < (3, 4):
        _safe_names = {'None': None, 'True': True, 'False': False}
    if isinstance(node_or_string, string_type):
        node_or_string = compile(
            node_or_string,
            '<string>' if file_name is None else file_name, 'eval', PyCF_ONLY_AST)
    if isinstance(node_or_string, Expression):
        node_or_string = node_or_string.body
    else:
        raise TypeError("only string or AST nodes supported")

    def _convert(node, expect_string=False):
        if isinstance(node, (Str, Bytes)):
            return node.s
        if expect_string:
            pass
        elif isinstance(node, Num):
            return node.n
        elif isinstance(node, Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, List):
            return list(map(_convert, node.elts))
        elif isinstance(node, Set):
            return set(map(_convert, node.elts))
        elif isinstance(node, Dict):
            return dict_typ((_convert(k, expect_string=True), _convert(v)) for k, v
                            in zip(node.keys, node.values))
        elif isinstance(node, NameConstant):
            return node.value
        elif sys.version_info < (3, 4) and isinstance(node, Name):
            if node.id in _safe_names:
                return _safe_names[node.id]
        elif isinstance(node, UnaryOp) and \
             isinstance(node.op, (UAdd, USub)) and \
             isinstance(node.operand, (Num, UnaryOp, BinOp)):  # NOQA
            operand = _convert(node.operand)
            if isinstance(node.op, UAdd):
                return + operand
            else:
                return - operand
        elif isinstance(node, BinOp) and \
             isinstance(node.op, (Add, Sub, Mult)) and \
             isinstance(node.right, (Num, UnaryOp, BinOp)) and \
             isinstance(node.left, (Num, UnaryOp, BinOp)):  # NOQA
            left = _convert(node.left)
            right = _convert(node.right)
            if isinstance(node.op, Add):
                return left + right
            elif isinstance(node.op, Mult):
                return left * right
            else:
                return left - right
        elif isinstance(node, Call):
            func_id = getattr(node.func, 'id', None)
            if func_id == 'dict':
                return dict_typ((k.arg, _convert(k.value)) for k in node.keywords)
            elif func_id == 'set':
                return set(_convert(node.args[0]))
            elif func_id == 'date':
                return datetime.date(*[_convert(k) for k in node.args])
            elif func_id == 'datetime':
                return datetime.datetime(*[_convert(k) for k in node.args])
            elif func_id == 'dedent':
                return dedent(*[_convert(k) for k in node.args])
        elif isinstance(node, Name):
            return node.s
        err = SyntaxError('malformed node or string: ' + repr(node))
        err.filename = '<string>'
        err.lineno = node.lineno
        err.offset = node.col_offset
        err.text = repr(node)
        err.node = node
        raise err
    res = _convert(node_or_string)
    if not isinstance(res, dict_typ):
        raise SyntaxError("Top level must be dict not " + repr(type(res)))
    if return_ast:
        return res, node_or_string
    return res


class Formatter(string.Formatter):
    def format(self, *args, **kwargs):
        self.nr_expanded = 0  # gets changed if a field is gotten
        return string.Formatter.format(self, *args, **kwargs)

    def get_field(self, field_name, args, kwargs):
        self.nr_expanded += 1
        return get(kwargs, field_name), field_name


class PON(object):
    sep = '.'
    MAX_DEPTH = 10

    def __init__(self, stream_or_list_or_string=None, obj=None):
        self.rt_info = [[], {}, [], [], False]
        if stream_or_list_or_string is not None:
            if isinstance(stream_or_list_or_string, string_type):
                stream_or_list_or_string = stream_or_list_or_string.splitlines(True)
            if obj is not None:
                raise NotImplementedError("cannot provide stream and object")
            self.extract(stream_or_list_or_string, start='', typ=None)
        else:
            self.obj = obj
        self.fmt = Formatter()

    def load(self, path):
        with open(path) as fp:
            txt = fp.read()
            if sys.version_info < (3,):
                txt = txt.decode('utf-8')
            return loads(txt)

    def get(self, path, expand=None):
        """path is either a string of sections seprated by PON.sep, or a list
        or a tuple (already completely split into sections).
        """
        if expand is True:
            expand = self.obj
        if isinstance(path, tuple):
            sections = list(path)
        elif isinstance(path, list):
            sections = path[:]
        else:
            sections = path.split(PON.sep)
        base = self.obj
        while sections:
            key = sections.pop(0)
            if isinstance(base, list):
                key = int(key)
            base = base[key]
        if expand is not None and isinstance(base, string_type):
            self.fmt.nr_expanded = -1
            depth = 0
            while self.fmt.nr_expanded != 0:
                base = self.fmt.format(base, **expand)
                depth += 1
                if depth > PON.MAX_DEPTH:
                    raise NotImplementedError("Max recursion depth is {0}".format(
                        PON.MAX_DEPTH))
            return base
        return base

    def store(self, path, value):
        sections = path.split(PON.sep)
        base = self.obj
        while len(sections) > 1:
            key = sections.pop(0)
            if isinstance(base, list):
                key = int(key)
            base = base[key]
        base[sections[0]] = value
        return self.obj

    def extract(self, stream_or_list_or_string, start, typ=' = dict(', round_trip=True):
        """read through the stream or list until the string start+typ is found.
        Then accumulate the configuration until the token typ[-1] is found
        in the same position as where start was.
        """

        if isinstance(stream_or_list_or_string, string_type):
            stream_or_list_or_string = stream_or_list_or_string.splitlines(True)
        d = ordereddict if round_trip else dict
        if typ is None:
            res = loads(u''.join(stream_or_list_or_string), dict_typ=d,
                        return_ast=round_trip)
        else:
            process = Process(start, typ)
            for line in stream_or_list_or_string:
                if process(line):
                    break
            res = loads(u''.join(process.lines), dict_typ=d, return_ast=round_trip)
            stream_or_list_or_string = process.lines
        if not round_trip:
            return res
        self.obj = res[0]
        if hasattr(stream_or_list_or_string, 'rewind'):
            stream_or_list_or_string.rewind()
        self.add_comments_to_keys(stream_or_list_or_string, res[1])

    def update(self, file_name, start, typ=' = dict('):
        """update the configuration in a file replacing the part start+typ until the endline
        of that structure. Normally used as

            file_name="__init__.py"
            s = "_package_data"
            t = ' = {'
            with open(file_name) as fp:
                p = PON()
                p.extract(fp, start=s, typ=t)
            # p.dictify()  # <- optional, e.g. if old format
            v = list(p.obj['version_info'])
            v[1] += 1 # update minor version
            v[2] = 0 # set micro version
            p.obj['version_info'] = tuple(v)
            p.update(file_name, start=s, typ=t)
        """
        import io
        with io.open(file_name) as fp:
            lines = list(fp)
        process = Process(start, typ)
        with io.open(file_name, 'w') as fp:
            for line in lines:
                res = process(line)
                if res is False:
                    print(line, end=u'', file=fp)
                elif res is True:
                    fp.write(u'{} = '.format(start))
                    self.dump(fp)

    def add_comments_to_keys(self, lines, ast):
        """ tokenize lines and gather comments
        comments are string, line, column, last_token_ended tuples.
        last_token_ended:
           0 -> full line
           -1 -> full line before dict started.
        """
        key_positions = self._parse_rt_info_from_ast(ast)
        comments = []
        dict_has_started = False
        last_token_ended = -1
        last_dict_line = 0
        for typ, token, start, end, l in tokenize.generate_tokens(ReadLiner(lines).readline):
            # tokenize.printtoken(typ, token, start, end, line)
            if typ == tokenize.COMMENT:
                comments.append((token, start[0], start[1], last_token_ended))
                continue
            if typ == tokenize.NL:
                last_token_ended = 0 if dict_has_started else -1
            else:
                last_token_ended = end[1]
                dict_has_started = True
                if typ != tokenize.ENDMARKER:
                    last_dict_line = start[0]
        ci = 0
        pre = []
        trailing = []
        post = []
        self.rt_info = [pre, key_positions, trailing, post, key_positions.pop(None, None)]
        while ci < len(comments) and comments[ci][3] < 0:
            pre.append(comments[ci])
            ci += 1
        for key in key_positions:
            v = key_positions[key]
            while ci < len(comments) and comments[ci][1] <= v[0]:
                if comments[ci][1] == v[0]:
                    v[4] = comments[ci]
                else:
                    v[3].append(comments[ci])
                ci += 1
        while ci < len(comments) and comments[ci][1] < last_dict_line:
            trailing.append(comments[ci])
            ci += 1
        while ci < len(comments):
            post.append(comments[ci])
            ci += 1

    def dump_rt_info(self):
        for c in self.rt_info[0]:
            print('pre', c)
        for key in self.rt_info[1]:
            v = self.rt_info[1][key]
            print(key, v)
        for c in self.rt_info[2]:
            print('trailing', c)
        for c in self.rt_info[3]:
            print('post', c)

    def _parse_rt_info_from_ast(self, a):
        key_positions = ordereddict()
        self.prefix = []
        res = self._convert(a, key_positions)
        if not isinstance(res, ordereddict):
            raise SyntaxError("Top level must be dict not " + repr(type(res)))
        return key_positions

    def _add_key(self, key_positions, key, line, col, value):
        """
        value == -3: dedented multi-line-string
        value == -2: curly braces style dict ( i.e.  '{}' not 'dict()' )
        value == -1: multi-line-string
        value >= 0: other value item, col_offset
        """
        self.prefix.append(key)
        key_positions[tuple(self.prefix)] = [line, col, value, [], None]

    def _curly_braces_style(self, key_positions):
        """we have to know the dict style ( {}/dict() )"""
        if not self.prefix:  # top level
            key_positions[None] = -2
        else:
            key_positions[tuple(self.prefix)][2] = -2

    def _dedented_multi_line_string(self, key_positions):
        """dedented multiline string"""
        key_positions[tuple(self.prefix)][2] = -3

    def _pop_key(self):
        self.prefix.pop()

    def _get_line_number(self, node):
        lineno = node.lineno
        if node.col_offset == -1:  # multiline string
            try:
                lineno -= node.s.count('\n')
            except:
                # change if anything else then multiline strings
                # can cause this to happen
                raise
        return lineno, node.col_offset, node.col_offset

    def _convert(self, node, key_positions):
        if isinstance(node, List):
            l = []
            for idx, v in enumerate(node.elts):
                line, col, mls = self._get_line_number(v)
                self._add_key(key_positions, idx, line, col, mls)
                l.append(self._convert(v, key_positions))
                self.prefix.pop()
            return l
        elif isinstance(node, Dict):
            d = ordereddict()
            for k, v in zip(node.keys, node.values):
                self._curly_braces_style(key_positions)
                # keys for curly braces style dict are nodes themselves, there is
                # no need for compensation
                kc = k.s  # Str/Bytes Node
                self._add_key(key_positions, kc, k.lineno, k.col_offset, v.col_offset)
                d[kc] = self._convert(v, key_positions)
                self._pop_key()
            return d
        elif isinstance(node, Call):
            func_id = getattr(node.func, 'id', None)
            if func_id == 'dict':
                d = ordereddict()
                for k in node.keywords:
                    line, col, mls = self._get_line_number(k.value)
                    self._add_key(key_positions, k.arg, line, col, mls)
                    d[k.arg] = self._convert(k.value, key_positions)
                    self._pop_key()
                return d
            elif func_id == 'set':
                return set(self._convert(node.args[0], key_positions))
            elif func_id == 'dedent':
                self._dedented_multi_line_string(key_positions)
                return None
        elif isinstance(node, Name):
            if sys.version_info < (3, 4) and isinstance(node, Name) and \
               node.id in ['None', 'True', 'False']:
                return node.id
            return node.s

    def dictify(self):
        """try to make "curly braces" style dicts in the config into "dict()" style ones"""
        import keyword    # to test string when dictify-ing, clashes with from ast import *

        def keys_ok_for_dict(iterable):  # hand in dict, or keys
            for sub_key in iterable:
                if not isinstance(sub_key, string_type):
                    return False
                if ' ' in sub_key:
                    return False
                if keyword.iskeyword(sub_key):
                    return False
            return True

        if keys_ok_for_dict(self.obj):
            self.rt_info[4] = False
        for k in self.rt_info[1]:
            v = self.rt_info[1][k]
            if len(v) > 1 and v[2] == -2:
                # check here for any subkeys with spaces or other "dict()"
                # prohibiting behaviour.
                if keys_ok_for_dict(self.get(k)):
                    v[2] = 1

    def dump(self, fp=sys.stdout):
        from pon.dump import PrettyPrinter

        pp = PrettyPrinter(stream=fp, indent=4, width=80, depth=None,
                           rt_info=self.rt_info)
        pp.pprint(self.obj)


def get(obj, path, expand=None):
    return PON(obj=obj).get(path, expand=expand)


def store(obj, path, value):
    return PON(obj=obj).store(path, value)


def extract(stream_or_list, start, typ=' = dict('):
    return PON().extract(stream_or_list, start, typ=typ, round_trip=False)


class ReadLiner:
    def __init__(self, lines):
        self.lines = lines
        self.idx = -1

    def readline(self):
        self.idx += 1
        if self.idx >= len(self.lines):
            raise StopIteration
        return self.lines[self.idx]


class Process:
    def __init__(self, start, typ):
        self.lines = []
        self.start = start
        self.typ = typ
        self.match = start + typ
        self.indent = None
        # don't have to do list in next match, top level should be dict
        self.end = {'{': '}', '(': ')', '[': ']'}[typ[-1]]

    def __call__(self, line):
        assert isinstance(line, string_type)
        if self.indent is None:
            pos = line.find(self.match)
            if pos == -1:
                return False
            self.indent = pos
            line = line.split('=', 1)[-1].lstrip()
        self.lines.append(line)
        if line[self.indent] == self.end:
            self.indent = None
            return True



# # header
# coding: utf-8

from __future__ import print_function

# # __init__.py parser

import sys
import os
import platform
from _ast import *       # NOQA
from ast import parse

from setuptools import setup, Extension, Distribution  # NOQA
from setuptools.command import install_lib

if __name__ != '__main__':
    raise NotImplementedError('should never include setup.py')

# # definitions

full_package_name = None

if __name__ != '__main__':
    raise NotImplementedError('should never include setup.py')

if sys.version_info < (3, ):
    string_type = basestring
else:
    string_type = str


if sys.version_info < (3, 4):
    class Bytes():
        pass

    class NameConstant:
        pass

if sys.version_info < (2, 7) or platform.python_implementation() == 'Jython':
    class Set():
        pass


def literal_eval(node_or_string):
    """
    Safely evaluate an expression node or a string containing a Python
    expression.  The string or node provided may only consist of the following
    Python literal structures: strings, bytes, numbers, tuples, lists, dicts,
    sets, booleans, and None.
    """
    _safe_names = {'None': None, 'True': True, 'False': False}
    if isinstance(node_or_string, string_type):
        node_or_string = parse(node_or_string, mode='eval')
    if isinstance(node_or_string, Expression):
        node_or_string = node_or_string.body
    else:
        raise TypeError("only string or AST nodes supported")

    def _convert(node):
        if isinstance(node, (Str, Bytes)):
            return node.s
        elif isinstance(node, Num):
            return node.n
        elif isinstance(node, Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, List):
            return list(map(_convert, node.elts))
        elif isinstance(node, Set):
            return set(map(_convert, node.elts))
        elif isinstance(node, Dict):
            return dict((_convert(k), _convert(v)) for k, v
                        in zip(node.keys, node.values))
        elif isinstance(node, NameConstant):
            return node.value
        elif sys.version_info < (3, 4) and isinstance(node, Name):
            if node.id in _safe_names:
                return _safe_names[node.id]
        elif isinstance(node, UnaryOp) and \
             isinstance(node.op, (UAdd, USub)) and \
             isinstance(node.operand, (Num, UnaryOp, BinOp)):  # NOQA
            operand = _convert(node.operand)
            if isinstance(node.op, UAdd):
                return + operand
            else:
                return - operand
        elif isinstance(node, BinOp) and \
             isinstance(node.op, (Add, Sub)) and \
             isinstance(node.right, (Num, UnaryOp, BinOp)) and \
             isinstance(node.left, (Num, UnaryOp, BinOp)):  # NOQA
            left = _convert(node.left)
            right = _convert(node.right)
            if isinstance(node.op, Add):
                return left + right
            else:
                return left - right
        elif isinstance(node, Call):
            func_id = getattr(node.func, 'id', None)
            if func_id == 'dict':
                return dict((k.arg, _convert(k.value)) for k in node.keywords)
            elif func_id == 'set':
                return set(_convert(node.args[0]))
            elif func_id == 'date':
                return datetime.date(*[_convert(k) for k in node.args])
            elif func_id == 'datetime':
                return datetime.datetime(*[_convert(k) for k in node.args])
        err = SyntaxError('malformed node or string: ' + repr(node))
        err.filename = '<string>'
        err.lineno = node.lineno
        err.offset = node.col_offset
        err.text = repr(node)
        err.node = node
        raise err
    return _convert(node_or_string)


# parses python ( "= dict( )" ) or ( "= {" )
def _package_data(fn):
    data = {}
    with open(fn) as fp:
        parsing = False
        lines = []
        for line in fp.readlines():
            if sys.version_info < (3,):
                line = line.decode('utf-8')
            if line.startswith(u'_package_data'):
                if 'dict(' in line:
                    parsing = 'python'
                    lines.append(u'dict(\n')
                elif line.endswith(u'= {\n'):
                    parsing = 'python'
                    lines.append(u'{\n')
                else:
                    raise NotImplementedError
                continue
            if not parsing:
                continue
            if parsing == 'python':
                if line.startswith(u')') or line.startswith(u'}'):
                    lines.append(line)
                    try:
                        data = literal_eval(u''.join(lines))
                    except SyntaxError as e:
                        context = 2
                        from_line = e.lineno - (context + 1)
                        to_line = e.lineno + (context - 1)
                        w = len(str(to_line))
                        for index, line in enumerate(lines):
                            if from_line <= index <= to_line:
                                print(u"{:{}}: {}".format(index, w, line), end=u'')
                                if index == e.lineno - 1:
                                    print(u"{:{}}  {}^--- {}".format(
                                        u' ', w, u' ' * e.offset, e.node))
                        raise
                    break
                lines.append(line)
            else:
                raise NotImplementedError
    return data

# make sure you can run "python ../some/dir/setup.py install"
pkg_data = _package_data(__file__.replace('setup.py', '__init__.py'))

exclude_files = [
    'setup.py',
]


# # helper
def _check_convert_version(tup):
    """Create a PEP 386 pseudo-format conformant string from tuple tup."""
    ret_val = str(tup[0])  # first is always digit
    next_sep = "."  # separator for next extension, can be "" or "."
    nr_digits = 0  # nr of adjacent digits in rest, to verify
    post_dev = False  # are we processig post/dev
    for x in tup[1:]:
        if isinstance(x, int):
            nr_digits += 1
            if nr_digits > 2:
                raise ValueError("too many consecutive digits after " + ret_val)
            ret_val += next_sep + str(x)
            next_sep = '.'
            continue
        first_letter = x[0].lower()
        next_sep = ''
        if first_letter in 'abcr':
            if post_dev:
                raise ValueError("release level specified after "
                                 "post/dev: " + x)
            nr_digits = 0
            ret_val += 'rc' if first_letter == 'r' else first_letter
        elif first_letter in 'pd':
            nr_digits = 1  # only one can follow
            post_dev = True
            ret_val += '.post' if first_letter == 'p' else '.dev'
        else:
            raise ValueError('First letter of "' + x + '" not recognised')
    # .dev and .post need a number otherwise setuptools normalizes and complains
    if nr_digits == 1 and post_dev:
        ret_val += '0'
    return ret_val


version_info = pkg_data['version_info']
version_str = _check_convert_version(version_info)


class MyInstallLib(install_lib.install_lib):
    def install(self):
        fpp = pkg_data['full_package_name'].split('.')  # full package path
        full_exclude_files = [os.path.join(*(fpp + [x]))
                              for x in exclude_files]
        alt_files = []
        outfiles = install_lib.install_lib.install(self)
        for x in outfiles:
            for full_exclude_file in full_exclude_files:
                if full_exclude_file in x:
                    os.remove(x)
                    break
            else:
                alt_files.append(x)
        return alt_files


class NameSpacePackager(object):
    def __init__(self, pkg_data):
        assert isinstance(pkg_data, dict)
        self._pkg_data = pkg_data
        self.full_package_name = self.pn(self._pkg_data['full_package_name'])
        self._split = None
        self.depth = self.full_package_name.count('.')
        self.command = None
        self._pkg = [None, None]  # required and pre-installable packages
        if sys.argv[0] == 'setup.py' and sys.argv[1] == 'install' and \
           '--single-version-externally-managed' not in sys.argv:
            print('error: you have to install with "pip install ."')
            sys.exit(1)
        # If you only support an extension module on Linux, Windows thinks it
        # is pure. That way you would get pure python .whl files that take
        # precedence for downloading on Linux over source with compilable C
        if self._pkg_data.get('universal'):
            Distribution.is_pure = lambda *args: True
        else:
            Distribution.is_pure = lambda *args: False
        for x in sys.argv:
            if x[0] == '-' or x == 'setup.py':
                continue
            self.command = x
            break

    def pn(self, s):
        if sys.version_info < (3, ) and isinstance(s, unicode):
            return s.encode('utf-8')
        return s

    @property
    def split(self):
        """split the full package name in list of compontents traditionally
        done by setuptools.find_packages. This routine skips any directories
        with __init__.py that start with "_" or ".", or contain a
        setup.py/tox.ini (indicating a subpackage)
        """
        if self._split is None:
            fpn = self.full_package_name.split('.')
            self._split = []
            while fpn:
                self._split.insert(0, '.'.join(fpn))
                fpn = fpn[:-1]
            for d in os.listdir('.'):
                if not os.path.isdir(d) or d == self._split[0] or d[0] in '._':
                    continue
                # prevent sub-packages in namespace from being included
                x = os.path.join(d, 'setup.py')
                if os.path.exists(x):
                    if not os.path.exists(os.path.join(d, 'tox.ini')):
                        print('\n>>>>> found "{0}" without tox.ini <<<<<\n'
                              ''.format(x))
                    continue
                x = os.path.join(d, '__init__.py')
                if os.path.exists(x):
                    self._split.append(self.full_package_name + '.' + d)
        return self._split

    @property
    def namespace_packages(self):
        return self.split[:self.depth]

    def namespace_directories(self, depth=None):
        """return list of directories where the namespace should be created /
        can be found
        """
        res = []
        for index, d in enumerate(self.split[:depth]):
            # toplevel gets a dot
            if index > 0:
                d = os.path.join(*d.split('.'))
            res.append('.' + d)
        return res

    @property
    def package_dir(self):
        d = {
            # don't specify empty dir, clashes with package_data spec
            self.full_package_name: '.',
        }
        if len(self.split) > 1:  # only if package namespace
            d[self.split[0]] = self.namespace_directories(1)[0]
        return d

    def create_dirs(self):
        """create the directories necessary for namespace packaging"""
        directories = self.namespace_directories(self.depth)
        if not directories:
            return
        if not os.path.exists(directories[0]):
            for d in directories:
                os.mkdir(d)
                with open(os.path.join(d, '__init__.py'), 'w') as fp:
                    fp.write('import pkg_resources\n'
                             'pkg_resources.declare_namespace(__name__)\n')

    def check(self):
        try:
            from pip.exceptions import InstallationError
        except ImportError:
            return
        # arg is either develop (pip install -e) or install
        if self.command not in ['install', 'develop']:
            return

        # if hgi and hgi.base are both in namespace_packages matching
        # against the top (hgi.) it suffices to find minus-e and non-minus-e
        # installed packages. As we don't know the order in namespace_packages
        # do some magic
        prefix = self.split[0]
        prefixes = set([prefix, prefix.replace('_', '-')])
        for p in sys.path:
            if not p:
                continue  # directory with setup.py
            if os.path.exists(os.path.join(p, 'setup.py')):
                continue  # some linked in stuff might not be hgi based
            if not os.path.isdir(p):
                continue
            if p.startswith('/tmp/'):
                continue
            for fn in os.listdir(p):
                for pre in prefixes:
                    if fn.startswith(pre):
                        break
                else:
                    continue
                full_name = os.path.join(p, fn)
                # not in prefixes the toplevel is never changed from _ to -
                if fn == prefix and os.path.isdir(full_name):
                    # directory -> other, non-minus-e, install
                    if self.command == 'develop':
                        raise InstallationError(
                            'Cannot mix develop (pip install -e),\nwith '
                            'non-develop installs for package name {0}'.format(
                                fn))
                elif fn == prefix:
                    raise InstallationError(
                        'non directory package {0} in {1}'.format(
                            fn, p))
                for pre in [x + '.' for x in prefixes]:
                    if fn.startswith(pre):
                        break
                else:
                    continue  # hgiabc instead of hgi.
                if fn.endswith('-link') and self.command == 'install':
                    raise InstallationError(
                        'Cannot mix non-develop with develop\n(pip install -e)'
                        ' installs for package name {0}'.format(fn))

    def entry_points(self, script_name=None, package_name=None):
        """normally called without explicit script_name and package name
        the default console_scripts entry depends on the existence of __main__.py:
        if that file exists then the function main() in there is used, otherwise
        the in __init__.py.

        the _package_data entry_points key/value pair can be explicitly specified
        including a "=" character. If the entry is True or 1 the
        scriptname is the last part of the full package path (split on '.')
        if the ep entry is a simple string without "=", that is assumed to be
        the name of the script.
        """
        def pckg_entry_point(name):
            return '{0}{1}:main'.format(
                name,
                '.__main__' if os.path.exists('__main__.py') else '',
            )

        ep = self._pkg_data.get('entry_points', True)
        if ep is None:
            return None
        if ep not in [True, 1]:
            if '=' in ep:
                # full specification of the entry point like
                # entry_points=['yaml = ruamel.yaml.cmd:main'],
                return {'console_scripts': [ep]}
            # assume that it is just the script name
            script_name = ep
        if package_name is None:
            package_name = self.full_package_name
        if not script_name:
            script_name = package_name.split('.')[-1]
        return {'console_scripts': [
            '{0} = {1}'.format(script_name, pckg_entry_point(package_name)),
        ]}

    @property
    def url(self):
        if self.full_package_name.startswith('ruamel.'):
            sp = self.full_package_name.split('.', 1)
        else:
            sp = ['ruamel', self.full_package_name]
        return 'https://bitbucket.org/{0}/{1}'.format(*sp)

    @property
    def author(self):
        return self._pkg_data['author']

    @property
    def author_email(self):
        return self._pkg_data['author_email']

    @property
    def license(self):
        """return the license field from _package_data, None means MIT"""
        lic = self._pkg_data.get('license')
        if lic is None:
            # lic_fn = os.path.join(os.path.dirname(__file__), 'LICENSE')
            # assert os.path.exists(lic_fn)
            return "MIT license"
        return lic

    def has_mit_lic(self):
        return 'MIT' in self.license

    @property
    def description(self):
        return self._pkg_data['description']

    @property
    def status(self):
        # αβ
        status = self._pkg_data.get('status', u'β')
        if status == u'α':
            return (3, 'Alpha')
        elif status == u'β':
            return (4, 'Beta')
        elif u'stable' in status.lower():
            return (5, 'Production/Stable')
        raise NotImplementedError

    @property
    def classifiers(self):
        return [
            'Development Status :: {0} - {1}'.format(*self.status),
            'Intended Audience :: Developers',
            'License :: ' + ('OSI Approved :: MIT' if self.has_mit_lic()
                             else 'Other/Proprietary') + ' License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
        ] + [self.pn(x) for x in self._pkg_data.get('classifiers', [])]

    @property
    def keywords(self):
        return self.pn(self._pkg_data.get('keywords'))

    @property
    def install_requires(self):
        """list of packages required for installation"""
        return self._analyse_packages[0]

    @property
    def install_pre(self):
        """list of packages required for installation"""
        return self._analyse_packages[1]

    @property
    def _analyse_packages(self):
        """gather from configuration, names starting with * need
        to be installed explicitly as they are not on PyPI
        install_requires should be  dict, with keys 'any', 'py27' etc
        or a list (which is as if only 'any' was defined
        """
        if self._pkg[0] is None:
            self._pkg[0] = []
            self._pkg[1] = []

        ir = self._pkg_data.get('install_requires')
        if ir is None:
            return self._pkg  # these will be both empty at this point
        if isinstance(ir, list):
            self._pkg[0] = ir
            return self._pkg
        # 'any' for all builds, 'py27' etc for specifics versions
        packages = ir.get('any', [])
        implementation = platform.python_implementation()
        if implementation == 'CPython':
            pyver = 'py{0}{1}'.format(*sys.version_info)
        elif implementation == 'PyPy':
            pyver = 'pypy' if sys.version_info < (3, ) else 'pypy3'
        elif implementation == 'Jython':
            pyver = 'jython'
        packages.extend(ir.get(pyver, []))
        for p in packages:
            # package name starting with * means use local source tree,  non-published
            # to PyPi or maybe not latest version on PyPI -> pre-install
            if p[0] == '*':
                p = p[1:]
                self._pkg[1].append(p)
            self._pkg[0].append(p)
        return self._pkg

    @property
    def data_files(self):
        df = self._pkg_data.get('data_files', [])
        if self.has_mit_lic():
            df.append('LICENSE')
        if not df:
            return None
        return [('.', df), ]

    @property
    def package_data(self):
        df = self._pkg_data.get('data_files', [])
        if self.has_mit_lic():
            # include the file
            df.append('LICENSE')
            # but don't install it
            exclude_files.append('LICENSE')
        if not df:
            return {}
        return {self.full_package_name: df}

    @property
    def ext_modules(self):
        """check if the C module can be build by trying to compile a small
            program against the libyaml development library"""
        if hasattr(self, '_ext_modules'):
            return self._ext_modules
        if '--version' in sys.argv:
            return None
        if platform.python_implementation() == 'Jython':
            return None
        if sys.platform == "win32" and not self._pkg_data.get('win32bin'):
            return None
        import tempfile
        import shutil
        from textwrap import dedent

        import distutils.sysconfig
        import distutils.ccompiler
        from distutils.errors import CompileError, LinkError

        self._ext_modules = []
        for target in self._pkg_data.get('ext_modules', []):  # list of dicts
            test_code = target.get('test')
            libraries = [self.pn(x) for x in target.get('lib')]
            ext = Extension(
                self.pn(target['name']),
                sources=[self.pn(x) for x in target['src']],
                libraries=libraries,
            )
            if not test_code:
                self._ext_modules.append(ext)
                continue
            # write a temporary .c file to compile
            c_code = dedent(target['test'])
            try:
                tmp_dir = tempfile.mkdtemp(prefix='tmp_ruamel_')
                bin_file_name = 'test' + self.pn(target['name'])
                file_name = os.path.join(tmp_dir, bin_file_name + '.c')
                with open(file_name, 'w') as fp:
                    fp.write(c_code)

                # and try to compile it
                compiler = distutils.ccompiler.new_compiler()
                assert isinstance(compiler, distutils.ccompiler.CCompiler)
                distutils.sysconfig.customize_compiler(compiler)

                try:
                    compiler.link_executable(
                        compiler.compile(
                            [file_name],
                            output_dir='/',  # as file_name has absolute prefix
                        ),
                        bin_file_name,
                        output_dir=tmp_dir,
                        libraries=libraries,
                    )
                except CompileError:
                    print('compile error:', file_name)
                    continue
                except LinkError:
                    print('libyaml link error', file_name)
                    continue
                self._ext_modules.append(ext)
            except Exception as e:  # NOQA
                # print('Exception:', e)
                pass
            finally:
                shutil.rmtree(tmp_dir)
        return self._ext_modules

    def wheel(self, kw, setup):
        """temporary add setup.cfg if creating a wheel to include LICENSE file
        https://bitbucket.org/pypa/wheel/issues/47
        """
        if 'bdist_wheel' not in sys.argv:
            return
        file_name = 'setup.cfg'
        if os.path.exists(file_name):  # add it if not in there?
            return
        with open(file_name, 'w') as fp:
            if os.path.exists('LICENSE'):
                fp.write('[metadata]\nlicense-file = LICENSE\n')
            else:
                print("\n\n>>>>>> LICENSE file not found <<<<<\n\n")
            if self._pkg_data.get('universal'):
                fp.write('[bdist_wheel]\nuniversal = 1\n')
        try:
            setup(**kw)
        except:
            raise
        finally:
            os.remove(file_name)
        return True


# # call setup
def main():
    nsp = NameSpacePackager(pkg_data)
    nsp.check()
    nsp.create_dirs()
    kw = dict(
        name=nsp.full_package_name,
        namespace_packages=nsp.namespace_packages,
        version=version_str,
        packages=nsp.split,
        url=nsp.url,
        author=nsp.author,
        author_email=nsp.author_email,
        cmdclass={'install_lib': MyInstallLib},
        package_dir=nsp.package_dir,
        entry_points=nsp.entry_points(),
        description=nsp.description,
        install_requires=nsp.install_requires,
        license=nsp.license,
        classifiers=nsp.classifiers,
        keywords=nsp.keywords,
        package_data=nsp.package_data,
        ext_modules=nsp.ext_modules,
    )
    if '--version' not in sys.argv and '--verbose' in sys.argv:
        for k in sorted(kw):
            v = kw[k]
            print(k, '->', v)
    with open('README.rst') as fp:
        kw['long_description'] = fp.read()
    if nsp.wheel(kw, setup):
        return
    for x in ['-c', 'egg_info', '--egg-base', 'pip-egg-info']:
        if x not in sys.argv:
            break
    else:
        # we're doing a tox setup install any starred package by searching up the source tree
        # until you match your/package/name for your.package.name
        for p in nsp.install_pre:
            import subprocess
            # search other source
            setup_path = os.path.join(*p.split('.') + ['setup.py'])
            try_dir = os.path.dirname(sys.executable)
            while len(try_dir) > 1:
                full_path_setup_py = os.path.join(try_dir, setup_path)
                if os.path.exists(full_path_setup_py):
                    pip = sys.executable.replace('python', 'pip')
                    cmd = [pip, 'install', os.path.dirname(full_path_setup_py)]
                    # with open('/var/tmp/notice', 'a') as fp:
                    #     print('installing', cmd, file=fp)
                    subprocess.check_output(cmd)
                    break
                try_dir = os.path.dirname(try_dir)
    setup(**kw)


main()



# coding: utf-8

import sys as _sys
import warnings

from pon import ordereddict, string_type


# try:
#     from cStringIO import StringIO as _StringIO
# except ImportError:
#     from io import StringIO as _StringIO
from io import StringIO as _StringIO

__all__ = ["PrettyPrinter"]

# cache these for faster access:
_commajoin = ", ".join
_id = id
_len = len
_type = type


def _sorted(iterable):
    with warnings.catch_warnings():
        if hasattr(_sys, 'py3kwarning') and _sys.py3kwarning:
            warnings.filterwarnings("ignore", "comparing unequal types "
                                    "not supported", DeprecationWarning)
        return sorted(iterable)


class PrettyPrinter:
    def __init__(self, indent=4, width=80, depth=None, stream=None, rt_info=None):
        """Handle pretty printing operations onto a stream using a set of
        configured parameters.

        indent
            Number of spaces to indent for each level of nesting.

        width
            Attempted maximum number of columns in the output.

        depth
            The maximum depth to print out nested structures.

        stream
            The desired output stream.  If omitted (or false), the standard
            output stream available at construction will be used.

        """
        indent = int(indent)
        width = int(width)
        assert indent >= 0, "indent must be >= 0"
        assert depth is None or depth > 0, "depth must be > 0"
        assert width, "width must be != 0"
        self._depth = depth
        self._indent_per_level = indent
        self._width = width
        if stream is not None:
            self._stream = stream
        else:
            self._stream = _sys.stdout
        self._rt_info = rt_info
        self._prefix = []
        self._eol = None
        self._col = 0
        self._same_line_no_info = -10

    def _add_key(self, key):
        self._prefix.append(key)

    def _pop_key(self):
        self._prefix.pop()

    def write(self, s, stream=None):
        if stream is None:
            stream = self._stream
        for ch in s:
            if ch == '\n':
                self._col = 0
                if self._eol:
                    stream.write(self._eol)
                    self._eol = None
                stream.write(ch)
                continue
            stream.write(ch)
            self._col += 1

    def decode_write(self, s, stream=None):
        if _sys.version_info < (3, ) and isinstance(s, str):
            s = s.decode('utf-8')
        self.write(s, stream=stream)

    def pprint(self, object):
        for c in self._rt_info[0]:
            self.write(u"{0}{1}\n".format(' ' * c[2], c[0]))
        self._format(object, self._stream, 0, 0, {}, 0)
        self.write(u"\n")
        for c in self._rt_info[3]:
            self.write(u"{0}{1}\n".format(' ' * c[2], c[0]))

    def pformat(self, object):
        sio = _StringIO()
        self._format(object, sio, 0, 0, {}, 0)
        return sio.getvalue()

    def indent(self, level):
        return u' ' * self._indent_per_level * level

    def rt_info_pre_eol(self):
        x = self._rt_info[1].get(tuple(self._prefix))
        if x is not None:
            return x[3:5]
        return [], None

    def rt_info_curly_braces(self):
        if not self._prefix:
            if len(self._rt_info) > 4:
                return self._rt_info[4]
            return False
        else:
            x = self._rt_info[1].get(tuple(self._prefix))
            if x is None:
                return False
            return x[2] == -2

    def rt_info_same_line(self):
        x = self._rt_info[1].get(tuple(self._prefix))
        if x is None:
            self._same_line_no_info -= 1
            return self._same_line_no_info
        return x[0]

    def _format(self, object, stream, indent, allowance, context, level):
        level = level + 1
        objid = _id(object)
        if objid in context:
            stream.write(_recursion(object))
            self._recursive = True
            self._readable = False
            return
        rep = self._repr(object, context, level - 1)
        typ = _type(object)
        sepLines = _len(rep) > (self._width - 1 - indent - allowance)
        write = self.write

        if self._depth and level > self._depth:
            write(rep)
            return

        r = getattr(typ, "__repr__", None)
        # print(type(object), typ == ordereddict, issubclass(typ, dict),
        #        r, ordereddict.__repr__)
        if issubclass(typ, dict) and \
           (r is dict.__repr__ or r == ordereddict.__repr__):
            curly_braces = self.rt_info_curly_braces()
            if not self._prefix:  # top level always separate lines
                sepLines = True
            else:
                sepLines = not curly_braces
            if curly_braces and not _len(object):
                write(u'{}')
                return
            if curly_braces:
                write(u'{')
                pre = None
                context[objid] = 1
                indent += 1
                items = list(object.items())
                idx = 0
                sep_lines = sepLines
                for key, ent in items:
                    idx += 1
                    self._add_key(key)
                    rep = self._repr(key, context, level)
                    pre, eol = self.rt_info_pre_eol()
                    for c in pre:
                        write(u"\n{0}{1}".format(' ' * c[2], c[0]))
                        sep_lines = True
                    if sep_lines:
                        write(u'\n%s%s: ' % (self.indent(indent), rep))
                    else:
                        write(u'%s: ' % rep)
                    if eol:
                        self._eol = "{0}{1}".format(' ' * (eol[2] - eol[3]), eol[0])
                    self._format(ent, stream, indent,  # + _len(rep) + 2,
                                 allowance + 1, context, level)
                    if sep_lines or idx != len(items):
                        write(u',')
                        if not sep_lines:
                            write(u' ')
                    self._pop_key()
                indent -= 1
                del context[objid]
                if sep_lines:
                    write(u'\n%s' % (self.indent(indent)))
                write(u'}')
                return
            else:
                write(u'dict(')
                length = _len(object)
                if length:
                    write(u'\n')
                    context[objid] = 1
                    indent += 1
                    items = object.items()
                    for key, ent in items:
                        rep = key
                        self._add_key(key)
                        pre, eol = self.rt_info_pre_eol()
                        for c in pre:
                            write(u"{0}{1}\n".format(' ' * c[2], c[0]))
                        if eol:
                            self._eol = u"{0}{1}".format(' ' * (eol[2] - eol[3]), eol[0])
                        if sepLines:
                            write(u'%s%s=' % (self.indent(indent), rep))
                        else:
                            write(u'%s=' % rep)
                        self._format(ent, stream, indent,
                                     allowance + 1, context, level)
                        self._pop_key()
                        write(u',\n')
                    indent -= 1
                    del context[objid]
                if not self._prefix:
                    for c in self._rt_info[2]:
                        write(u"{0}{1}\n".format(' ' * c[2], c[0]))
                write(u"%s%s" % (self.indent(indent), ')'))
                return

        if ((issubclass(typ, list) and r is list.__repr__) or
           (issubclass(typ, tuple) and r is tuple.__repr__) or
           (issubclass(typ, set) and r is set.__repr__) or
           (issubclass(typ, frozenset) and r is frozenset.__repr__)):
            length = _len(object)
            if issubclass(typ, list):
                write(u'[')
                endchar = ']'
            elif issubclass(typ, tuple):
                write(u'(')
                endchar = ')'
            else:
                if not length:
                    write(rep)
                    return
                self.decode_write(typ.__name__)
                write(u'([')
                endchar = '])'
                # indent += len(typ.__name__) + 1
                object = _sorted(object)
            end_sep = sepLines
            # if self._indent_per_level > 1 and sepLines:
            #     write((self._indent_per_level - 1) * ' ')
            idx = -1
            if length:
                same_line = self.rt_info_same_line()
                context[objid] = 1
                if sepLines:
                    indent += 1
                # self._add_key(idx)
                #  self._format(object[0], stream, indent, allowance + 1,
                #              context, level)
                # self._pop_key()
                if length > 0:
                    for ent in object:
                        idx += 1
                        self._add_key(idx)
                        if issubclass(typ, list):
                            my_line = self.rt_info_same_line()
                            # if we have no line info, then you get negative values -> one line
                            sep_lines = my_line != same_line if my_line >= 0 else False
                            same_line == my_line
                            end_sep = sep_lines
                        else:
                            sep_lines = sepLines
                        if idx:
                            if sep_lines:
                                write(u',\n')
                            else:
                                write(u', ')
                        else:  # first element
                            if sep_lines:
                                write(u'\n')
                        if issubclass(typ, list):
                            pre, eol = self.rt_info_pre_eol()
                            for c in pre:
                                write(u"{0}{1}\n".format(' ' * c[2], c[0]))
                            if eol:
                                self._eol = "{0}{1}".format(' ' * (eol[2] - eol[3]), eol[0])
                        if sep_lines:
                            self.write(self.indent(indent))
                        self._format(ent, stream, indent,
                                     allowance + 1, context, level)
                        self._pop_key()
                if sepLines:
                    indent -= 1
                del context[objid]
            if issubclass(typ, tuple) and length == 1:
                write(u',')
            if end_sep:
                write(u",\n%s%s" % (self.indent(indent), endchar))
            else:
                write(u"%s" % endchar)
            return
        if rep.startswith('datetime.date'):
            rep = rep.replace('datetime.date', 'date', 1)
            self.decode_write(rep)
            return
        if isinstance(rep, string_type):
            self.decode_write(rep)
            return
        print('unknown typ:', rep)
        raise NotImplementedError

    def _repr(self, object, context, level):
        repr1, readable, recursive = self.format(object, context.copy(),
                                                 self._depth, level)
        if not readable:
            self._readable = False
        if recursive:
            self._recursive = True
        return repr1

    def format(self, object, context, maxlevels, level):
        """Format object for a specific context, returning a string
        and flags indicating whether the representation is 'readable'
        and whether the object represents a recursive construct.
        """
        return self._safe_repr(object, context, maxlevels, level)

    # Return triple (repr_string, isreadable, isrecursive).

    def _safe_repr(self, object, context, maxlevels, level):
        typ = _type(object)
        if typ is str:
            str_typ = self._rt_info[1].get(tuple(self._prefix), [None, None, None])[2]
            # if str_typ not in [-1, -3] and 'locale' not in _sys.modules:
            #     rep = repr(object)
            #     if _sys.version_info < (3, 0):
            #         rep = rep.decode('utf-8')
            #     print('safe', typ, object)
            #     return rep, True, False
            pref = u''
            if str_typ in [-1, -3]:  # mls, dedented
                quotes = {'\n': '\n'}
            elif "'" in object and '"' not in object:
                closure = '"'
                quotes = {'"': '\\"'}
            else:
                closure = "'"
                quotes = {"'": "\\'"}
            qget = quotes.get
            sio = _StringIO()
            for char in object:
                if char.isalpha():
                    self.decode_write(char, stream=sio)
                    if _sys.version_info >= (3, ) and ord(char) > 255:
                        pref = u'u'
                else:
                    self.decode_write(qget(char, repr(char)[1:-1]), stream=sio)
            if str_typ == -3:  # dedented
                s = sio.getvalue().split('\n')
                s = (u'\n' + self.indent(level + 1)).join(s[:-1]) + \
                    u'\n' + self.indent(level) + s[-1]
                return (u'dedent("""%s""")' % (s)), True, False
            elif str_typ == -1:
                s = sio.getvalue()
                return (u'"""%s"""' % (s)), True, False
            return (u"%s%s%s%s" % (pref, closure, sio.getvalue(), closure)), True, False

        r = getattr(typ, "__repr__", None)
        # if issubclass(typ, dict) and r is dict.__repr__:
        if issubclass(typ, dict) and \
           (r is dict.__repr__ or r == ordereddict.__repr__):
            if not object:
                return u"{}", True, False
            objid = _id(object)
            if maxlevels and level >= maxlevels:
                return u"{...}", False, objid in context
            if objid in context:
                return _recursion(object), False, True
            context[objid] = 1
            readable = True
            recursive = False
            components = []
            append = components.append
            level += 1
            for k, v in object.items():
                krepr, kreadable, krecur = self._safe_repr(k, context, maxlevels, level)
                vrepr, vreadable, vrecur = self._safe_repr(v, context, maxlevels, level)
                append("%s: %s" % (krepr, vrepr))
                readable = readable and kreadable and vreadable
                if krecur or vrecur:
                    recursive = True
            del context[objid]
            return u"{%s}" % _commajoin(components), readable, recursive

        if (issubclass(typ, list) and r is list.__repr__) or \
           (issubclass(typ, tuple) and r is tuple.__repr__):
            if issubclass(typ, list):
                if not object:
                    return "[]", True, False
                format = u"[%s]"
            elif _len(object) == 1:
                format = u"(%s,)"
            else:
                if not object:
                    return u"()", True, False
                format = u"(%s)"
            objid = _id(object)
            if maxlevels and level >= maxlevels:
                return format % u"...", False, objid in context
            if objid in context:
                return _recursion(object), False, True
            context[objid] = 1
            readable = True
            recursive = False
            components = []
            append = components.append
            level += 1
            for o in object:
                orepr, oreadable, orecur = self._safe_repr(o, context, maxlevels, level)
                append(orepr)
                if not oreadable:
                    readable = False
                if orecur:
                    recursive = True
            del context[objid]
            return format % _commajoin(components), readable, recursive

        if _sys.version_info < (3,) and issubclass(typ, unicode):
            rep = u"u'{0}'".format(object)
        else:
            rep = repr(object)
        return rep, (rep and not rep.startswith(u'<')), False


def _recursion(object):
    return ("<Recursion on %s with id=%s>"
            % (_type(object).__name__, _id(object)))
